@@ -44,17 +44,18 @@ enum
     feed_name_offset = 1,
 };
 
-typedef enum entry_kind entry_kind_t;
-typedef enum file_status file_status_t;
-typedef enum flag_idx flag_idx_t;
+typedef enum entry_kind_t entry_kind_t;
+typedef enum file_status_t file_status_t;
+typedef enum flag_idx_t flag_idx_t;
 
-typedef struct revision revision_t;
-typedef struct filemap filemap_t;
-typedef struct feed feed_t;
-typedef struct entry entry_t;
+typedef struct revision_t revision_t;
+typedef struct filemap_t filemap_t;
+typedef struct err_t err_t;
+typedef struct feed_t feed_t;
+typedef struct entry_t entry_t;
 
 /* this enueration matters for backward compatibility */
-enum entry_kind
+enum entry_kind_t
 {
     ek_none = 0,
     ek_feed_item,
@@ -66,40 +67,48 @@ enum entry_kind
     ek_items_count,
 };
 
-enum file_status
+enum file_status_t
 {
     fs_unknown = 0,
     fs_noent,           // file does not exist yet
     fs_regular,         // file exists
     fs_unrecognized,    // file exists, format not recognized
     fs_incompat,        // file exists, recognized, different version
+    fs_corrupt,         // file exists, recognized, same version, corrupt
 };
 
 /* this enueration matters for backward compatibility */
-enum flag_idx
+enum flag_idx_t
 {
     fi_output_mode,
     fi_url_type,
     fi_flags_length,    // guard
 };
 
-struct revision
+struct revision_t
 {
     uint8_t major;
     uint8_t minor;
 } __attribute__((packed, aligned(1)));
 
-struct filemap
+struct filemap_t
 {
     void *memory;
     size_t size;
 };
 
-struct feed
+struct err_t
+{
+    /*  >= 0    errno_val = -value  error = sfe_system_error
+     *  < 0     errno_val = 0       error = value
+     */
+    int value;
+};
+
+struct feed_t
 {
     UT_hash_handle hh;
     const char *name;
-    savefile_t *save_ref;
     filemap_t filemap;
     struct {
         str_t provided;
@@ -111,24 +120,22 @@ struct feed
     file_status_t file_status;
     unsigned n_items;
     unsigned items_count;
+    int dirfd;          /* not owned. */
     uint8_t flags[fi_flags_length];
     bool marked_for_deletion : 1;
     bool needs_write : 1;
 };
 
-struct save
+struct savefile_t
 {
     int dirfd;
-    struct {
-        savefile_err_t kind;
-        int errno_val;
-    } last_error;
+    err_t last_error;
     feed_t *feeds;
     bool basedir_scanned : 1;
     bool feed_addition : 1;
 };
 
-struct entry
+struct entry_t
 {
     uint8_t kind;
     uint16_t length;
@@ -140,18 +147,23 @@ const char *file_status_str[] = {
     [fs_regular]        = "fs_regular",
     [fs_unrecognized]   = "fs_unrecognized",
     [fs_incompat]       = "fs_incompat",
+    [fs_corrupt]        = "fs_corrupt",
 };
 
 static const char *dotstrdup(const char *);
 static int writev_wrap(int fd, struct iovec *, size_t);
 static int filesize(int fd, size_t *s);
+
 static int filemap_load(filemap_t *, int fd);
 static void filemap_unload(filemap_t *);
 
+static err_t err_set(int enum_value);
+static savefile_err_t err_get(err_t);
+
 static int entry_write(int fd, const str_t *entry, entry_kind_t);
 static int entry_load(feed_t *, int offset, str_t *, entry_kind_t *);
 
-static feed_t *feed_new(savefile_t *, const char *fname);
+static feed_t *feed_new(int dirfd, const char *fname);
 static const char * feed_name(const feed_t *);
 static bool feed_can_be_written(const feed_t *);
 static int feed_load_magic(feed_t *);
@@ -160,16 +172,15 @@ static int feed_load(feed_t *, int fd);
 static int feed_write_magic(int fd);
 static int feed_write_fields(const feed_t *, int fd);
 static int feed_write_items_count(const feed_t *, int fd);
-static int feed_write(const feed_t *, int basedir);
-static int feed_unlink(feed_t *, int basedir);
-static int feed_update(feed_t *, int basedir);
+static int feed_write(feed_t *);
+static int feed_unlink(feed_t *);
 static void feed_set_str_field(feed_t *f, str_t *field, str_t *new_value);
 static void feed_set_flag(feed_t *f, flag_idx_t idx, uint8_t byte);
 static void feed_flag_to_write(feed_t *f);
-static void feed_del(feed_t *);
+static void feed_free_fields(feed_t *);
+static void feed_free(feed_t *);
 static int feed_add_item(feed_t *, str_t *);
 
-static void save_fail(savefile_t *, savefile_err_t);
 static feed_t *save_find_feed(savefile_t *, const char *fname);
 static bool save_is_bogus_fname(const char *fname);
 static feed_t *save_add_feed(savefile_t *save, const char *fname);
@@ -240,8 +251,10 @@ static int filemap_load(filemap_t *fmap, int fd)
         goto success;
 
     mapped = mmap(NULL, size, PROT_READ, MAP_PRIVATE, fd, 0);
-    if (mapped == MAP_FAILED)
+    if (mapped == MAP_FAILED) {
+        warn("mmap failed");
         return -1;
+    }
 
     *fmap = (filemap_t) {
         .memory = mapped,
@@ -262,6 +275,26 @@ static void filemap_unload(filemap_t *fmap)
     *fmap = (filemap_t) {};
 }
 
+static err_t err_set(int enum_value)
+{
+    return (err_t){
+        .value = (enum_value == sfe_system_error) ? -errno : enum_value
+    };
+}
+
+static savefile_err_t err_get(err_t err)
+{
+    if (err.value >= 0)
+        return (savefile_err_t){
+            .error = err.value
+        };
+
+    return (savefile_err_t){
+        .error = sfe_system_error,
+        .errno_val = -err.value,
+    };
+}
+
 static int entry_write(int fd, const str_t *str, entry_kind_t ek)
 {
     entry_t entry = {
@@ -278,6 +311,14 @@ static int entry_write(int fd, const str_t *str, entry_kind_t ek)
     return writev_wrap(fd, iovec, sizeof(iovec));
 }
 
+static int flag_corrupt(feed_t *feed)
+{
+    warnx("feed [%s]: file too short", feed_name(feed));
+    test_counter(flag_corrupt);
+    feed->file_status = fs_corrupt;
+    return 0;
+}
+
 static int entry_load(feed_t *feed,
                       int read_offset,
                       str_t *str,
@@ -286,13 +327,13 @@ static int entry_load(feed_t *feed,
     entry_t *entry;
 
     if (sizeof(entry_t) > feed->filemap.size - read_offset)
-        goto bad;
+        return flag_corrupt(feed);
 
     entry = (entry_t *)((uint8_t *)feed->filemap.memory + read_offset);
     read_offset += sizeof(entry_t);
 
     if (entry->length > feed->filemap.size - read_offset)
-        goto bad;
+        return flag_corrupt(feed);
 
     *ek = entry->kind;
     *str = (str_t){
@@ -302,22 +343,24 @@ static int entry_load(feed_t *feed,
     read_offset += entry->length;
 
     return read_offset;
-
-bad:
-    warnx("feed [%s]: file too short", feed_name(feed));
-    feed->file_status = fs_unrecognized;
-    return 0;
 }
 
-static feed_t *feed_new(savefile_t *save, const char *fname)
+static feed_t *feed_new(int dirfd, const char *fname)
 {
-    int fd = openat(save->dirfd, fname, O_RDONLY);
+    feed_t *feed = NULL;
+    int fd;
+
+    /* We try to open the feed file.   It might not exist, but this is not
+     * a problem, as we might create one later.  We fail in case openat
+     * fails for reason other than ENOENT.
+     */
+    fd = openat(dirfd, fname, O_RDONLY | O_CLOEXEC);
     if (fd == -1 && errno != ENOENT) {
-        warn("openat(%d, %s, O_RDONLY)", save->dirfd, fname);
-        return NULL;
+        warn("openat(%d, %s, O_RDONLY)", dirfd, fname);
+        goto fail;
     }
 
-    feed_t *feed = malloc(sizeof(feed_t));
+    feed = malloc(sizeof(feed_t));
     if (!feed) {
         warn("malloc failed");
         goto fail;
@@ -325,7 +368,7 @@ static feed_t *feed_new(savefile_t *save, const char *fname)
 
     *feed = (feed_t) {
         .name = dotstrdup(fname),
-        .save_ref = save,
+        .dirfd = dirfd,
     };
 
     if (!feed->name)
@@ -333,14 +376,25 @@ static feed_t *feed_new(savefile_t *save, const char *fname)
 
     if (fd == -1)
         feed->file_status = fs_noent;
-    else if (feed_load(feed, fd) == -1)
-        goto fail;
+    else {
+        if (feed_load(feed, fd) == -1)
+            goto fail;
+
+        /* A corrupt file can be used and written over, but we have to be
+         * sure to get rid of what we managed to load from the corrupt
+         * file. */
+        if (feed->file_status == fs_corrupt) {
+            feed_free_fields(feed);
+            filemap_unload(&feed->filemap);
+        }
+    }
 
     return feed;
 
 fail:
-    feed_del(feed);
-    close(fd);
+    feed_free(feed);
+    if (fd != -1)
+        close(fd);
     return NULL;
 }
 
@@ -356,6 +410,7 @@ static bool feed_can_be_written(const feed_t *feed)
         assert(false);  // is only used during initialization
     case fs_noent:
     case fs_regular:
+    case fs_corrupt:
         return true;
     case fs_unrecognized:
     case fs_incompat:
@@ -366,16 +421,25 @@ static bool feed_can_be_written(const feed_t *feed)
 
 static int feed_load_magic(feed_t *feed)
 {
+    /* Check if the magic number exists.
+     *
+     * returns -1 on failure (it is expected by contract, but it never
+     *      happens in reality.
+     * returns 0 if the file is unrecognized.
+     * returns the offset of the first byte beyond the magic number if the
+     *      file is recognized.
+     */
+
     if (feed->filemap.size < sizeof(magic_nr) + sizeof(revision_t)) {
         whisper("feed [%s]: no magic number", feed_name(feed));
-        goto bad;
+        goto unrecognized;
     }
 
     const uint8_t *bytes = feed->filemap.memory;
 
     if (strncmp(magic_nr, (const char *)bytes, sizeof(magic_nr)) != 0) {
         whisper("feed [%s]: invalid magic number", feed_name(feed));
-        goto bad;
+        goto unrecognized;
     }
     bytes += sizeof(magic_nr);
 
@@ -392,7 +456,7 @@ static int feed_load_magic(feed_t *feed)
 
     return bytes - (uint8_t *)feed->filemap.memory;
 
-bad:
+unrecognized:
     feed->file_status = fs_unrecognized;
     return 0;
 }
@@ -409,6 +473,14 @@ static void feed_load_items_count(feed_t *feed, const str_t *repr)
 
 static int feed_load(feed_t *feed, int fd)
 {
+    /* This function is only invoked when constructing the feed.  Failing
+     * here (return -1) implies that the 'feed' object will be destroyed,
+     * therefore assigning 'last_error' is not required.  In failure
+     * situations which are not related to system errors (e.g. corrupt
+     * file) this function is going to be successful, but the
+     * 'file_status' field will be set accordingly.
+     */
+
     if (filemap_load(&feed->filemap, fd) == -1)
         return -1;
 
@@ -417,18 +489,19 @@ static int feed_load(feed_t *feed, int fd)
         return read_offset;
 
     while (read_offset < feed->filemap.size) {
-        entry_kind_t ek;
-        int add_offset;
+        entry_kind_t ek = ek_none;
+        int updated_offset;
         str_t content;
 
-        add_offset = entry_load(feed, read_offset, &content, &ek);
-        if (add_offset <= 0)
-            return add_offset;
+        updated_offset = entry_load(feed, read_offset, &content, &ek);
+        if (updated_offset <= 0)
+            return updated_offset;
 
         switch (ek) {
         case ek_feed_item:
             /* failure explicitly ignored */
-            (void)feed_add_item(feed, &content);
+            if (feed_add_item(feed, &content) == -1)
+                goto done;
             break;
 
         case ek_url_provided:
@@ -464,15 +537,19 @@ static int feed_load(feed_t *feed, int fd)
             break;
 
         case ek_none:
+            /* Most likely a bug.  Probably in entry_load? */
+            warnx("feed [%s]: reports entry of kind ek_none??",
+                  feed_name(feed));
         default:
             whisper("feed [%s]: ignoring entry kind %d",
                     feed_name(feed), ek);
             break;
         }
 
-        read_offset = add_offset;
+        read_offset = updated_offset;
     }
 
+done:
     feed->file_status = fs_regular;
     return 0;
 }
@@ -537,7 +614,7 @@ static int feed_write_items_count(const feed_t *feed, int fd)
     return writev_wrap(fd, iovec, sizeof(iovec));
 }
 
-static int feed_write(const feed_t *feed, int basedir)
+static int feed_write(feed_t *feed)
 {
     /* this function saves the feed on a file with the same name as the
      * original feed file, but prefixed with a dot.  If the write is
@@ -551,9 +628,9 @@ static int feed_write(const feed_t *feed, int basedir)
      * of this memory optimization.
      */
 
-    int fd = openat(basedir, feed->name,
-        O_WRONLY | O_APPEND | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR |
-        S_IRGRP | S_IWGRP
+    int fd = openat(feed->dirfd, feed->name,
+        O_WRONLY | O_APPEND | O_CREAT | O_EXCL | O_CLOEXEC,
+        S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP
     );
 
     if (fd == -1) {
@@ -585,25 +662,30 @@ static int feed_write(const feed_t *feed, int basedir)
         return -1;
     }
 
-    if (renameat(basedir, feed->name, basedir, feed_name(feed)) == -1)
+    if (renameat(feed->dirfd, feed->name, feed->dirfd, feed_name(feed)) == -1)
         return -1;
 
     test_counter(feed_written);
+    feed->needs_write = false;
     return 0;
 }
 
-static int feed_unlink(feed_t *feed, int basedir)
+static int feed_unlink(feed_t *feed)
 {
     const char *filename = feed_name(feed);
 
-    if (unlinkat(basedir, filename, 0) == 0)
+    if (unlinkat(feed->dirfd, filename, 0) == 0) {
+        test_counter(feed_unlinked);
+        feed_flag_to_write(feed);
+        feed->marked_for_deletion = false;
         return 0;
+    }
 
-    warn("unlikat(%d, %s) failed", basedir, filename);
+    warn("unlikat(%d, %s) failed", feed->dirfd, filename);
     return -1;
 }
 
-static int feed_update(feed_t *feed, int basedir)
+int feed_persist(feed_t *feed)
 {
     if (!feed_can_be_written(feed)) {
         warnx(
@@ -614,10 +696,10 @@ static int feed_update(feed_t *feed, int basedir)
     }
 
     if (feed->marked_for_deletion)
-        return feed_unlink(feed, basedir);
+        return feed_unlink(feed);
 
     if (feed->needs_write)
-        return feed_write(feed, basedir);
+        return feed_write(feed);
 
     return 0;
 }
@@ -646,35 +728,45 @@ static void feed_flag_to_write(feed_t *feed)
     feed->needs_write = true;
 }
 
-static void feed_clear(feed_t *feed)
+static void feed_clear_items(feed_t *feed)
 {
     for (unsigned i = 0; i < feed->n_items; ++i)
         str_free(&feed->items[i]);
     feed->n_items = 0;
 }
 
-static void feed_del(feed_t *feed)
+static void feed_free_fields(feed_t *feed)
+{
+    /* All the fields freed in this function have in common the fact that
+     * they might be pointers to 'feed->filemap', unless differently
+     * assigned by means of setters.
+     */
+    str_free(&feed->url.provided);
+    str_free(&feed->url.effective);
+    str_free(&feed->outfmt);
+    str_free(&feed->subproc_chdir);
+    feed_clear_items(feed);
+}
+
+static void feed_free(feed_t *feed)
 {
     if (!feed)
         return;
 
-    filemap_unload(&feed->filemap);
     free((void *)feed->name);
 
-    str_free(&feed->url.provided);
-    str_free(&feed->url.effective);
-    str_free(&feed->outfmt);
-
-    feed_clear(feed);
+    feed_free_fields(feed);
+    filemap_unload(&feed->filemap);
     free(feed);
 }
 
 static int feed_add_item(feed_t *feed, str_t *data)
 {
     if (feed->n_items >= feed_max_items) {
-        save_fail(feed->save_ref, sfe_max_resources);
-        warnx("feed [%s]: ignoring item %.*s: no space left",
-              feed_name(feed), STR_FMT(data));
+        warnx(
+            "feed_add_item: maximum number of feed items (%d) exceeded",
+            feed_max_items
+        );
         return -1;
     }
 
@@ -690,16 +782,22 @@ static int save_scan_feeds(savefile_t *save)
 
     test_counter(save_scan_called);
 
+    /* NOTE: here we dup the descriptor so that our copy is not closed
+     * when we invoke closedir.  The iteration over directory entries will
+     * however affect the offset of 'save->dirfd'.  This is OK, since
+     * save_scan_feeds is protected from a second execution, but if it
+     * wasn't so, the second iteration would yield no result. */
     int dirfd = dup(save->dirfd);
+
     if (dirfd == -1) {
-        save_fail(save, sfe_system_error);
+        save->last_error = err_set(sfe_system_error);
         warn("dup failed");
         goto exit;
     }
 
     dir = fdopendir(dirfd);
     if (!dir) {
-        save_fail(save, sfe_system_error);
+        save->last_error = err_set(sfe_system_error);
         warn("fdopendir failed");
         goto exit;
     }
@@ -711,7 +809,7 @@ static int save_scan_feeds(savefile_t *save)
         struct stat sb;
 
         if (fstatat(save->dirfd, dent->d_name, &sb, 0) == -1) {
-            save_fail(save, sfe_system_error);
+            save->last_error = err_set(sfe_system_error);
             warn("fstatat(%d, %s) failed", save->dirfd, dent->d_name);
             goto exit;
         }
@@ -749,13 +847,6 @@ exit:
     return e;
 }
 
-static void save_fail(savefile_t *save, savefile_err_t e)
-{
-    save->last_error.kind = e;
-    if (e == sfe_system_error)
-        save->last_error.errno_val = errno;
-}
-
 static feed_t *save_find_feed(savefile_t *save, const char *fname)
 {
     feed_t *feed;
@@ -792,13 +883,13 @@ static feed_t *save_add_feed(savefile_t *save, const char *fname)
     feed_t *feed;
 
     if (save_is_bogus_fname(fname)) {
-        save_fail(save, sfe_bad_input);
+        save->last_error = err_set(sfe_bad_input);
         return NULL;
     }
 
-    feed = feed_new(save, fname);
+    feed = feed_new(save->dirfd, fname);
     if (!feed) {
-        save_fail(save, sfe_system_error);
+        save->last_error = err_set(sfe_system_error);
         return NULL;
     }
 
@@ -838,16 +929,12 @@ savefile_t * savefile_new(int basedir)
 
 savefile_err_t savefile_get_error(const savefile_t *save)
 {
-    savefile_err_t e;
-    e = save->last_error.kind;
-    if (e == sfe_system_error)
-        errno = save->last_error.errno_val;
-    return e;
+    return err_get(save->last_error);
 }
 
 void savefile_reset_error(savefile_t *save)
 {
-    save->last_error.kind = sfe_none;
+    save->last_error = err_set(sfe_none);
 }
 
 feed_t *savefile_get_feed(savefile_t *save, const char *fname)
@@ -864,8 +951,8 @@ feed_t *savefile_get_feed(savefile_t *save, const char *fname)
     if (feed_can_be_written(feed))
         return feed;
 
-    save_fail(save, sfe_protected_file);
-    return NULL;
+    save->last_error = err_set(sfe_protected_file);
+    return NULL;    /* note: feed is owned by savefile */
 }
 
 int savefile_load_all(savefile_t *save)
@@ -935,12 +1022,11 @@ int feed_set_subproc_chdir(feed_t *feed, str_t *path)
 void feed_mark_deleted(feed_t *feed)
 {
     feed->marked_for_deletion = true;
-    feed_flag_to_write(feed);
 }
 
 int feed_set_items(feed_t *feed, const feed_items_t *new_items)
 {
-    feed_clear(feed);
+    feed_clear_items(feed);
 
     for (unsigned i = 0; i < new_items->n_items; ++i) {
         str_t item = {};
@@ -968,9 +1054,9 @@ str_t feed_get_effective_url(const feed_t *feed)
     return STR_ALIAS(&feed->url.effective);
 }
 
-str_t feed_get_name(const feed_t *feed)
+const char *feed_get_name(const feed_t *feed)
 {
-    return STR_CAST(feed_name(feed));
+    return feed_name(feed);
 }
 
 str_t feed_get_outfmt(const feed_t *feed)
@@ -1046,28 +1132,20 @@ feed_url_type_t feed_get_url_type(const feed_t *feed)
     return feed->flags[fi_url_type];
 }
 
-int savefile_free(savefile_t *save, bool write_changes)
+void savefile_free(savefile_t *save)
 {
-    int e = 0;
     feed_t *feed, *tmp;
 
     if (!save)
-        return e;
-
-    int dirfd = save->dirfd;
+        return;
 
     HASH_ITER(hh, save->feeds, feed, tmp) {
         HASH_DEL(save->feeds, feed);
-
-        if (write_changes && feed_update(feed, dirfd) == -1)
-            e = -1;
-
-        feed_del(feed);
+        feed_free(feed);
     }
 
-    if (dirfd != -1)
-        close(dirfd);
+    if (save->dirfd != -1)
+        close(save->dirfd);
 
     free(save);
-    return e;
 }
